#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/device.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/drivers/gpio.h>
#include <stdio.h>
#include "OBSL-ADS131M0x/SPI_IO_nRF.h"
#include "OBSL-ADS131M0x/ads131m0x.h"
#include "OBSL-ADS131M0x/ads131_reg_defs.h"

#define SPIOP_ADS131 (SPI_WORD_SET(8) | SPI_TRANSFER_MSB | SPI_MODE_CPHA)
struct spi_dt_spec spi_dev = SPI_DT_SPEC_GET(DT_NODELABEL(ads131m02), SPIOP_ADS131, 0);

#define ADS_NODE DT_NODELABEL(ads131m02)
const struct gpio_dt_spec drdy_gpio = GPIO_DT_SPEC_GET(ADS_NODE, drdy_gpios);
static struct gpio_callback drdy_cb;

volatile bool interrupt_flag = false;
volatile bool busy = false;

#define NUM_SAMPLES 1024
int sample_index = 0;

uint32_t start_time = 0;
uint32_t end_time = 0;

void drdy_isr_handler(const struct device *port, struct gpio_callback *cb, uint32_t pins)
{
    ARG_UNUSED(port);
    ARG_UNUSED(cb);
    ARG_UNUSED(pins);
    interrupt_flag = true;
}

int main(void)
{
    if (!spi_is_ready_dt(&spi_dev)) {
        printk("SPI not ready\n");
        return 0;
    }

    if (!device_is_ready(drdy_gpio.port)) {
        printk("DRDY GPIO not ready\n");
        return 0;
    }

    // Configure DRDY GPIO and ISR
    gpio_pin_configure_dt(&drdy_gpio, GPIO_INPUT);
    gpio_pin_interrupt_configure_dt(&drdy_gpio, GPIO_INT_EDGE_TO_ACTIVE);
    gpio_init_callback(&drdy_cb, drdy_isr_handler, BIT(drdy_gpio.pin));
    gpio_add_callback(drdy_gpio.port, &drdy_cb);

    // Read the device ID
    uint16_t id = 0;
    if (ADS131_GetID(&spi_dev, &id) == ADS131M0x_OK) {
        printk("Device ID: 0x%04X\n", id);
    } else {
        printk("Failed to read ID\n");
    }

    // device is out of reset
    ADS131_Mode_SetReset(&spi_dev, ADS131M0x_MODE_RESET_CLEAR_BIT);

    // set OSR
    ADS131_Clock_SetOSR(&spi_dev, ADS131M0x_CLOCK_OSR_128);
    k_msleep(10);
    uint16_t clk_reg = 0;

    ADS131M0x_ReadRegister(&spi_dev, ADS131M0x_CLOCK, &clk_reg);
    printk("CLOCK Reg Confirm: 0x%04X\n", clk_reg);

    // clear pending DRDY
    while (gpio_pin_get_dt(&drdy_gpio) == 0) {
        k_busy_wait(10);
    }

    interrupt_flag = false;
    sample_index = 0;
    start_time = k_uptime_get();

    while (sample_index < NUM_SAMPLES) {
        if (interrupt_flag && !busy) {
            busy = true;
            interrupt_flag = false;

            uint8_t tx_buf[15] = {0};
            uint8_t rx_buf[15] = {0};

            if (SPI_IO_Transfer(&spi_dev, tx_buf, rx_buf, sizeof(tx_buf)) != 0) {
                printk("SPI transfer failed\n");
                busy = false;
                continue;
            }

            int32_t raw0 = (rx_buf[3] << 16) | (rx_buf[4] << 8) | rx_buf[5];
            int32_t raw1 = (rx_buf[6] << 16) | (rx_buf[7] << 8) | rx_buf[8];

            // Sign extension
            if (raw0 & 0x800000) raw0 |= 0xFF000000;
            if (raw1 & 0x800000) raw1 |= 0xFF000000;

            // printf("%d %d\n", raw0, raw1);

            sample_index++;
            busy = false;
        }
    }

    end_time = k_uptime_get();
    uint32_t duration = end_time - start_time;
    float sps = (1000.0f * NUM_SAMPLES) / duration;

    printk("\nSampling Complete\n");
    printk("Duration: %u ms\n", duration);
    printk("Effective SPS: %d\n", (int)sps);

    while (1) k_sleep(K_FOREVER);
    return 0;
}
